
// (c) COPYRIGHT URI/MIT 1996, 1999
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      Dan Holloway <d.holloway@gso.uri.edu> dan

// This file contains functions used to build the DAS and DDS objects
// returned by the jg_das and jg_dds filter programs.

#include "config_jg.h"

static char rcsid[] not_used ={"$Id: jg_util.cc,v 1.3 2003/01/16 23:11:48 jimg Exp $"};

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pwd.h>
#include <sys/stat.h>

#include <string>

#include "DDS.h"
#include "DAS.h"
#include "util.h"
#include "cgi_util.h"
#include "dods-limits.h"

#include "jgofs.h"      // extern declarations including many globals.

#include "JGByte.h"
#include "JGInt16.h"
#include "JGUInt16.h"
#include "JGInt32.h"
#include "JGUInt32.h"
#include "JGFloat32.h"
#include "JGFloat64.h"
#include "JGSequence.h"
#include "JGStr.h"

using std::string;

#define HOME_DIR_HACK 1

// Note that these globals are defined here and declared extern in `jgofs.h'.
// Including that file gives you access to these variables and is required
// when using the JGOFS code. jhrg 12/31/97.

char varName[NVALS][TOKEN];
char varValue[NVALS][VALSIZE];
int varLevel[NVALS];
int nVars, unit;
int varSize = TOKEN;

// Could this use string::operator==()? 01/15/03 jhrg
static AttrType
get_attribute_type(const string &return_attr) 
{
    if (return_attr.find("container") != return_attr.npos)
	return Attr_container;
    else if (return_attr.find("byte") != return_attr.npos)
	return Attr_byte;
    else if (return_attr.find("uint16") != return_attr.npos)
	return Attr_uint16;
    else if (return_attr.find("int16") != return_attr.npos)
	return Attr_int16;
    else if (return_attr.find("uint32") != return_attr.npos)
	return Attr_uint32;
    else if (return_attr.find("int32") != return_attr.npos)
	return Attr_int32;
    else if (return_attr.find("float32") != return_attr.npos)
	return Attr_float32;
    else if (return_attr.find("float64") != return_attr.npos)
	return Attr_float64;
    else if (return_attr.find("string") != return_attr.npos)
	return Attr_string;
    else if (return_attr.find("url") != return_attr.npos)
	return Attr_url;
    else 
	return Attr_unknown;
}

static inline string
build_error_msg(const string &var, const string &attr, const string &type)
{
    return string("\"JGOFS/DDS - Variable = ") + var 
	+ string(", DODS_Type attr = ") + attr 
	+ string(", if provided in ancillary DAS file, the missing_value attribute must be type ") 
	+ type + string("\"");
}


// Check for an DAS container named 'DODS_Type' use the variable type
// specified in this container to convert JGOFS string variables to this
// type (where applicable).
//  
// If this container does not exist, and for those variables not listed
// in it, set the default variable type to 'string'.

static void
CreateSequence(int currLvl, int maxLvls, DAS &das, JGSequence *seq)
    throw(Error)
{
    if (currLvl <= maxLvls) {
	// Create a new 'Sequence' variable representation, then create a
	// 'BaseType' representation for each variable at this level and add
	// it to current Sequence. When done call CreateSequence with
	// currLvl++.
	string name;
	name = "Level_" + long_to_string(currLvl);

	for (int i = 0; i < nVars; i++) {
	    if (varLevel[i] == currLvl) {

		AttrTable *at = das.get_table(varName[i]);

		if ( at ) {
		    string return_attr = at->get_attr("DODS_Type");
		    downcase(return_attr);
		    string dods_var = varName[i];

		    switch (get_attribute_type(return_attr)) {

		      case Attr_byte: {
			  JGByte *new_byte = new JGByte(varName[i]);

			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      // No Missing_Value specification, use default
			      // value.
			      new_byte->set_missing_value(DODS_UCHAR_MAX);
			  }
			  else if (var_miss == Attr_byte) {
			      new_byte->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      // Error condition, unless unspecified or 'byte'
			      // cannot set missing value to one specified in
			      // the ancillary DAS.
			      throw Error(build_error_msg(dods_var, return_attr,
							  "Byte"));
			  }

			  seq->add_var(new_byte);
			  break;
		      }
		    
		      case Attr_int16: {
			  JGInt16 *new_int16 = new JGInt16(varName[i]);

			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_int16->set_missing_value(DODS_SHRT_MAX);
			  }
			  else if ( var_miss == Attr_int16 ) {
			      new_int16->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "Int16"));
			  }

			  seq->add_var(new_int16);
			  break;
		      }

		      case Attr_uint16: {
			  JGUInt16 *new_uint16 = new JGUInt16(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_uint16->set_missing_value(DODS_USHRT_MAX);
			  }
			  else if ( var_miss == Attr_uint16 ) {
			      new_uint16->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "UInt16"));
			  }

			  seq->add_var(new_uint16);
			  break;
		      }

		      case Attr_int32: {
			  JGInt32 *new_int32 = new JGInt32(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_int32->set_missing_value(DODS_INT_MAX);
			  }
			  else if ( var_miss == Attr_int32 ) {
			      new_int32->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "Int32"));
			  }

			  seq->add_var(new_int32);
			  break;
		      }

		      case Attr_uint32: {
			  JGUInt32 *new_uint32 = new JGUInt32(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_uint32->set_missing_value(DODS_UINT_MAX);
			  }
			  else if ( var_miss == Attr_uint32 ) {
			      new_uint32->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "UInt32"));
			  }

			  seq->add_var(new_uint32);
			  break;
		      }

		      case Attr_float32: {
			  JGFloat32 *new_flt32 = new JGFloat32(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_flt32->set_missing_value(DODS_FLT_MAX);
			  }
			  else if ( var_miss == Attr_float32 ) {
			      new_flt32->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "Float32"));
			  }

			  seq->add_var(new_flt32);
			  break;
		      }

		      case Attr_float64: {
			  JGFloat64 *new_flt64 = new JGFloat64(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_flt64->set_missing_value(DODS_DBL_MAX);
			  }
			  else if ( var_miss == Attr_float64 ) {
			      new_flt64->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "Float64"));
			  }

			  seq->add_var(new_flt64);
			  break;
		      }

		      case Attr_string: {
			  JGStr *new_str = new JGStr(varName[i]);
			  AttrType var_miss = at->get_attr_type("missing_value");
			  if ( var_miss == Attr_unknown ) {
			      new_str->set_missing_value(string("nd"));
			  }
			  else if ( var_miss == Attr_string ) {
			      new_str->set_missing_value(at->get_attr("missing_value"));
			  }
			  else {
			      throw Error(build_error_msg(dods_var, return_attr,
							  "String"));
			  }

			  seq->add_var(new_str);
			  break;
		      }

		      case Attr_url: {
			case Attr_container:
			  string error;
			  error += "\"JGOFS/DDS - Cannot convert variable `" 
			      + (string)varName[i] + "' to type: " 
			      + at->get_attr(varName[i]) 
			      + ", valid types are Byte, Int16, UInt16, Int32, UInt32, Float32, Float64\"";		
			  throw Error(error);
		      }

		      case Attr_unknown: {
			  JGStr *new_str = new JGStr(varName[i]);

			  new_str->set_missing_value("nd");

			  seq->add_var(new_str);
			  break;
		      }
		    }
		}
		else { 
		    seq->add_var(new JGStr(varName[i]));
		}
	    }
	}
	
	CreateSequence(currLvl+1, maxLvls, das, seq); 
    }
}

// Given the jgofs variable id and a pointer to an attribute table pointer,
// read the variables attributes and store their names and values in the
// attribute table.
//
// This code handles the missing_value attribtue specially. 

static void
read_attributes(int v,  AttrTable *at) throw(Error)
{
    string return_attr = at->get_attr("DODS_Type");
    downcase(return_attr);

    char attributeStr[TOKEN];
    string dods_var = varName[v];
    
    switch (get_attribute_type(return_attr)) {
      case Attr_byte: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      // No Missing_Value specification, use default value.
	      unsigned char missing_byte_value = DODS_UCHAR_MAX;
	      sprintf(attributeStr, "%d", missing_byte_value);
	      at->append_attr("missing_value", "Byte", attributeStr);
	  }
	  else if ( var_miss != Attr_byte ) {
	      // Error condition, unless unspecified or 'byte' cannot set
	      // missing value to one specified in the ancillary DAS.
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type Byte for this variable.\"");
	  }
	  break;
      }

      case Attr_int16: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      short missing_int16_value = DODS_SHRT_MAX;
	      sprintf(attributeStr, "%d", missing_int16_value);
	      at->append_attr("missing_value", "Int16", attributeStr);
	  }
	  else if ( var_miss != Attr_int16 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type Int16 for this variable.\"");
	  }
	  break;
      }

      case Attr_uint16: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      unsigned short missing_uint16_value = DODS_USHRT_MAX;
	      sprintf(attributeStr, "%u", missing_uint16_value);
	      at->append_attr("missing_value", "UInt16", attributeStr);
	  }
	  else if ( var_miss != Attr_uint16 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type UInt16 for this variable.\"");
	  }
	  break;
      }

      case Attr_int32: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      long missing_int32_value = DODS_INT_MAX;
	      sprintf(attributeStr, "%ld", missing_int32_value);
	      at->append_attr("missing_value", "Int32", attributeStr);
	  }
	  else if ( var_miss != Attr_int32 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type Int32 for this variable.\"");
	  }
	  break;
      }

      case Attr_uint32: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      unsigned long missing_uint32_value = DODS_UINT_MAX;
	      sprintf(attributeStr, "%lu", missing_uint32_value);
	      at->append_attr("missing_value", "UInt32", attributeStr);
	  }
	  else if ( var_miss != Attr_uint32 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type UInt32 for this variable.\"");
	  }
	  break;
      }

      case Attr_float32: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      float missing_float32_value = DODS_FLT_MAX;
	      sprintf(attributeStr, "%e", missing_float32_value);
	      at->append_attr("missing_value", "Float32", attributeStr);
	  }
	  else if ( var_miss != Attr_float32 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type Float32 for this variable.\"");
	  }
	  break;
      }
      
      case Attr_float64: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      double missing_float64_value = DODS_DBL_MAX;
	      sprintf(attributeStr, "%e", missing_float64_value);
	      at->append_attr("missing_value", "Float64", attributeStr);
	  }
	  else if ( var_miss != Attr_float64 ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type Float64 for this variable.\"");
	  }
	  break;
      }
     
      case Attr_string: {
	  AttrType var_miss = at->get_attr_type("missing_value");
	  if ( var_miss == Attr_unknown ) {
	      at->append_attr("missing_value", "String", "\"nd\"");
	  }
	  else if ( var_miss != Attr_string ) {
	      throw Error(string("\"JGOFS/DAS - Variable = ") + dods_var 
			  + ", DODS_Type attr = " + return_attr 
			  + ", if provided in ancillary DAS file, missing_value attribute must be type String for this variable.\"");
	  }
	  break;
      }

      case Attr_url: 
      case Attr_container:
      case Attr_unknown: 
	// What about these? 01/15/03 jhrg
	break;

    }

    // Read attributes from the JGOFS dataset itself. Attributes may have
    // three forms: name=value, name-value or value. In the last case the
    // value is assumed to be the name of the units of the variable, that is,
    // the attribute's name is `units.'
    char attrname[TOKEN];
    while( ljdbattributes_(&unit, &v, attrname) ) {

	string attr = attrname;
	string::size_type idx;
	if (((idx = attr.find("=")) != string::npos) 
	    || ((idx = attr.find("-")) != string::npos)) {
	    string name = attr.substr(0,idx);
	    string val = attr.substr(idx+1, attr.length()-idx+1);

	    at->append_attr(name, "String", string("\"") + val + string("\""));
	}
	else {
	    at->append_attr("units", "String", 
			    string("\"") + attr + string("\""));
	}
    }
}

int 
test_path(const char *pathname)
{
    struct stat buf;

    if (lstat(pathname, &buf) == 0 
	&& (S_ISDIR(buf.st_mode) || S_ISLNK(buf.st_mode)))
	return 1;
    else
	return 0;
}

void
check_jgofs_path(string &relativePath, string &methodPath)
{

    /* The HOME_DIR_HACK uses the current machine's /etc/passwd file (or
       whatever getpwnam(3) searches) to find the JGOFS home directory. We
       assume that the jgofs object root directory is a subdir of the home
       dir called `objects'. This seems to be OK to do since the Makefile
       uses an equivalent assumption when it builds the JGOFS API. 10/13/95
       jhrg */

    /* HOME_DIR_HACK modified to search for the environment variables
       `JGOFS_OBJECT', and `JGOFS_METHOD' to satisfy the JGOFS dictionary
       search mechanism without requiring an entry in the /etc/passwd file.
       3/1/99 danh */

    string new_object_path = "", new_method_path = "";
    char *object_path, OBJECT_PATH[128];
    char *method_path, METHOD_PATH[128];

#if HOME_DIR_HACK

    struct passwd *pwd_entry;
    bool object_error = true, method_error = true;
    string errMessage = "";

    /* No `jgofs' user, OBJECT_PATH, METHOD_PATH must be defined, else exit
       with error */
    object_path = getenv("JGOFS_OBJECT");
    if ( object_path ) {
	object_error = false;

	new_object_path = object_path + relativePath;   /* append relative
							   directory */ 
	relativePath = new_object_path;   /* append relative directory */
	if (test_path(new_object_path.c_str()) == 0) {
	    object_error = TRUE;
	}
	else {
	    method_path = getenv("JGOFS_METHOD");
	    if ( method_path ) {
		method_error = false;

		new_method_path = method_path;
		methodPath = new_method_path;   /* append relative directory */

		if (test_path(new_method_path.c_str()) == 0) {
		    method_error = true;
		}
	    }
	    else {  /* METHOD_PATH is undefined */
		method_error = true;
	    }
	}
    }
    else {
	pwd_entry = getpwnam( "jgofs" );	/* assume the name "jgofs" */
	if ( pwd_entry == NULL ) {
	    /* OBJECT_PATH is undefined && No `jgofs' user defined. */
	    /* check to see if there is a '.objects' file in the    */
	    /* current directory.                                   */

	    struct stat buf;

	    if ( lstat(".objects",&buf) < 0 ) object_error = TRUE;
	    else {
		if ( S_ISREG(buf.st_mode)) {
		    object_error = false;

		    if ( (method_path = getcwd(METHOD_PATH,128)) ) {
			strcat(method_path,"/");
			method_error = false;
		    }
		    else 
			method_error = true;
		}
	    }
	}
	else {  
	
	    /* `jgofs' user exists, define OBJECT_PATH, METHOD_PATH relative
	       to home directory */
	    object_path = OBJECT_PATH;
	    strcpy(object_path, pwd_entry->pw_dir);

	    /* assume all dictionaries are in subdirs of `objects' */
	    strcat(object_path, "/objects");

	    if ( relativePath != "") { 
		strcat(object_path, "/");                  
		/* append relative_path directory */
		strcat(object_path, relativePath.c_str());
	    }
	    object_error = FALSE;

	    if (test_path(object_path) == 0) 
		object_error = true;
	
	    /* update relative_path to complete path */
	    relativePath = object_path;

	    method_path = METHOD_PATH;

	    strcpy(method_path, pwd_entry->pw_dir);
	    /* assume methods are in subdir `methods' */
	    strcat(method_path, "/methods/");

	    method_error = false;

	    if (test_path(method_path) == 0) 
		method_error = true;

	    methodPath = method_path;
	}
    }

    /* Check for errors in the OBJECT_PATH and METHOD_PATH locations. */
    if ( object_error || method_error ) {
	errMessage = "unable to determine JGOFS_OBJECT or JGOFS_METHOD, refer to the INSTALL file for the JGOFS server.";

	ErrMsgT(errMessage);

	throw InternalErr(errMessage);
    }

#else /* ! HOME_DIR_HACK */

    /* HOME_DIR_HACK is not set */
    object_path = OBJECT_PATH;
    strcpy(object_path, OBJECT_ROOT);	/* original objects */
    strcat(object_path, relativePath.c_str()); /* append relative directory */

    object_error = false;

    if (test_path(object_path) == 0)
	object_error = true;

    method_path = METHOD_PATH;
    strcpy(method_path, METHOD_ROOT);	/* original method  */

    method_error = false;

    if (test_path(method_path) == 0) 
	method_error = true;
#endif /* HOME_DIR_HACK */

}

// Given a reference to an instance of class DAS and a filename that referrs
// to a JGOFS object, read the JGOFS object and extract all the attributes
// of each of its variables. Add the variables and their attributes to the
// instance of DAS.

void
create_das(DAS &das, string &obj_path, string &met_path, string &jg_name)
    throw(Error)
{
    // Replaced -99 with -NVALS. jhrg 11/8/96
    nVars = -(NVALS);
    int nLvls = ljdbopen_(&unit, obj_path.c_str(), met_path.c_str(), 
			  jg_name.c_str(), (char*)varName, &varSize, &nVars);
    if (nLvls < 0) {
	char msg[1024];

        if (ljdberror_(&unit, static_cast<char *>(msg))) 
	    throw Error(string("\"Could not open ") + jg_name + ": " 
			+ msg + "\"");
	else 
	    throw Error(string("\"JGOFS/DDS - could not open object: ") 
			+  jg_name + "\"");
    }

    // `v' is the variable id passed to read_attributes().
    for (int v = 0; v < nVars; ++v) {
	AttrTable *attr_table_ptr = das.get_table(varName[v]);

	if (!attr_table_ptr)
	    attr_table_ptr = das.add_table(varName[v], new AttrTable);
      
	read_attributes(v, attr_table_ptr); // throws Error
    }
}

void
create_dds(DDS &dds, const string &object_path, const string &method_path, 
	   const string &jg_name, DAS &das) throw(Error)
{
    // Changed -99 to -NVALS (250). jhrg 11/8/96
    nVars = -(NVALS);
    int nLvls = ljdbopen_(&unit, object_path.c_str(), method_path.c_str(), 
			  jg_name.c_str(), (char *)varName, &varSize, &nVars);

    if (nLvls < 0) {
	string error;
	char msg[1024];

        if (ljdberror_(&unit, msg)) 
	    error += "\"Could not open " + jg_name + ": " + msg + "\"";
	else 
	    error += "\"JGOFS/DDS - could not open object: " +  jg_name + "\"";

	throw Error(error);
    }
    
    for (int v = 0; v < nVars; ++v) 
	varLevel[v] = ljdblevel_(&unit, &v);

    // When setting the dataset name, use only the part before the
    // constraint. 
    string::size_type pos;
    if ((pos = jg_name.find("(")) != string::npos)
	dds.set_dataset_name(name_path(jg_name.substr(0, pos)));
    else
	dds.set_dataset_name(name_path(jg_name));

    JGSequence *seq = new JGSequence("Level_0");
    
    CreateSequence(0, nLvls, das, seq);

    dds.add_var(seq);
}

// $Log: jg_util.cc,v $
// Revision 1.3  2003/01/16 23:11:48  jimg
// Updated the builds scripts to autoconf 2.53.
// Fixed warnings from gcc -Wall.
//
// Revision 1.2  2003/01/16 07:57:35  jimg
// Merged changes from release-3-2-6.
// Removed a number of files used for the client library ans sample clients.
//
// Revision 1.1.2.2  2003/01/16 01:31:15  jimg
// Removed some remaining old code.
//
// Revision 1.1.2.1  2003/01/16 01:30:26  jimg
// Added. This file was created by merging jgdas.cc, jgdds.cc and jg_path.cc.
// That code originally dates from late 1994. Over the years it has undergone
// considerable change. Use CVS to look at the logs of those files for
// information about the changes. In creating this file, I changed the error
// reporting scheme of the main functions to exceptions and made the code
// compatible with changes in the local jgofs API (in ljgofs). In the ldct.c
// and ljdb.c files I made many of the parameters const so that it was
// obvious which were passed by reference but did not have their values
// changed. This meant that several calls that could use string's c_str()
// method and avoid using intermediary char arrays.
//
